/* --- Copyright University of Birmingham 1998. All rights reserved. ------
 > File:            $poplocal/local/prb/auto/define_rulesystem.p
 > Purpose:         Implements define :rulesystem format
 > Author:          Aaron Sloman, Apr 17 1996 (see revisions)
 > Documentation:
 > Related Files:
 */

/*

define :rulesystem <name>
	[DLOCAL <var spec>]
    cycle_limit = <integer> ;		;;; optional
    [LVARS <popmatchvar spec>];		;;; optional
	debug = <boolean>;				;;; optional

    include: <rulefamily name>
    cycle_limit = <integer> ;		;;; optional
    include: <ruleset name>
    include: <rulefamily name>
			with_limit = <integer>	;;; optional
    include: <ruleset name>
		with_limit = <integer> ;	;;; optional
    ......

    ....

enddefine;

;;; TEST EXAMPLES

define :ruleset rs1 ;
	rule rs11 [a][b] ==> [c]
	rule rs12 [x] [y] ==> [c] [d]
enddefine;

rs1 ==>

define :ruleset rs2;
	rule rs21 [a][b] ==> [c]
	rule fred22 [x] [y] ==> [c] [d]
enddefine;

rs2 ==>

define :rulefamily vars rf1;

	[LVARS x [[y z] = 24//7] [w =99]];
	debug = true;
	use_section = true;

	ruleset: rs1
	ruleset: rs2
enddefine;

rf1.datalist ==>

vars q = conspair(3,4);

define :ruleset rs3;
	[VARS x [[y z] = destpair(q)] w];

	rule rs31 [a][b] ==> [c]

	rule rs32 [x] [y] ==> [c] [d]
enddefine;

rs3 ==>

define :ruleset rs4;
	rule rs41 [a][b] ==> [c]

	rule rs42 [x] [y] ==> [c] [d]
	rule rs43 [x y] ==> [do rs4] [c d]
enddefine;

rs4=>


define :rulefamily rf2;
	ruleset: rs3
	ruleset: rs4
enddefine;

rf2 ==>

define :rulesystem testrs;
	[DLOCAL [prb_walk = true] [prb_allrules = true]];
	cycle_limit = 4;
	[LVARS sim_myself [[myx myy]=sim_coords(sim_myself)]];

	include: rf1
	cycle_limit = 4;
	include: rs3
	include: rf2 with_limit = 3
enddefine;


testrs ==>


*/

uses poprulebase

section;

define lconstant plant_cycle_limit(item, rulesysname);
	if isword(item) then
		;;; should be a variable whose value is an integer.
		sysPUSH(item);
		sysPUSHQ(0);
		sysPUSHQ(false);
		sysCALL("fi_check");	;;; check integer >= 0
	elseif isinteger(item) then
		sysPUSHQ(item)
	else
		mishap('INTEGER NEEDED FOR CYCLE LIMIT',
			[^item rulesystem ^rulesysname])
	endif;
enddefine;

define :define_form global rulesystem;

	lvars
		rulesysname, dlocal_spec, familyname, item, cycle_limit,
		vars_spec, lvars_spec,
		debug, use_section;

	 prb_read_header(false)
		-> (rulesysname, dlocal_spec, cycle_limit, vars_spec, lvars_spec, debug, use_section);
	unless rulesysname then
		mishap('Name missing in "define :rulesystem"', [])
	endunless;

	if vars_spec then
		mishap(rulesysname, 1, '[VARS ...] NOT ALLOWED IN RULESYSTEM')
	endif;

	;;; plant call of conslist, after reading in rulefamily names, etc.

	lvars
		numrulefamilies = 0;

	if dlocal_spec and listlength(dlocal_spec) > 1 then
		;;; allow empty [DLOCAL
		sysPUSHQ(dlocal_spec);
		numrulefamilies + 1 -> numrulefamilies;
	endif;

	if cycle_limit then
		;;; The rulesystem will contain an integer
		plant_cycle_limit(cycle_limit, rulesysname);
		numrulefamilies + 1 -> numrulefamilies;
	endif;


	;;; spec for default popmatchvars
	if lvars_spec then
		sysPUSHQ(lvars_spec);
		numrulefamilies + 1 -> numrulefamilies;
	endif;

	repeat
		;;; pop11_need_nextreaditem([include enddefine ]) -> item;
		readitem() -> item;
		quitif(item == "enddefine");
		if item == "include" then
			;;; read colon, then rulefamily or ruleset name.
			pop11_need_nextreaditem(":") ->;
			readitem() -> item;		;;; rulesystem or rulefamily name
			;;; it's a ruleset
			sysPUSHQ(item);	;;; put the word in the list
			numrulefamilies + 1 -> numrulefamilies;

			lvars with_limit = false;
			if hd(proglist) == "with_limit" then
				readitem() ->;
				pop11_need_nextreaditem("=") ->;
				readitem() -> cycle_limit;
				plant_cycle_limit(cycle_limit, rulesysname);
				;;; read surplus semi-colon
				if hd(proglist) == ";" then
					readitem() ->
				endif;
				true -> with_limit;
			endif;

			if with_limit then
				sysCALL("conspair")
			endif;
		elseif item == "cycle_limit" then
			pop11_need_nextreaditem("=") ->;
			readitem() -> item;
			plant_cycle_limit(item, rulesysname);
			;;; read surplus semi-colon if present
			if hd(proglist) == ";" then
				readitem() ->
			endif;
			numrulefamilies + 1 -> numrulefamilies;
		else
			mishap('Expecting "include" or "enddefine" in define :rulesystem', [^item])
			;;; no other options yet
		endif;
	endrepeat;

	;;; create a list of all the rulefamilies, etc.
	sysPUSHQ(numrulefamilies);
	sysCALL("conslist");
	sysPOP(rulesysname);

enddefine;

endsection;

/* --- Revision History ---------------------------------------------------
--- Aaron Sloman, Sep 27 1998
	Allowed empty [DLOCAL] to simplify commenting out trace control
	expressions.
--- Aaron Sloman, May 26 1996
	Made semi colon after "with_limit = .." or "cycle_limit = ..." optional
--- Aaron Sloman,  25 May 1996
	Allowed local limits represented by a pair, consisting of ruleset or rulefamily
	and integer.
--- Aaron Sloman, May 17 1996
	Made it include rulefamilies, and made copies of them.
--- Aaron Sloman, May 17 1996
	Fixed bug that stopped names of rulesets/families being put in the
	list.
 */
